import logging
import os
import yaml
from logging.handlers import RotatingFileHandler


def send_stdout(message):
    print (message)


def load_imap_credentials(file='.acmagent'):
    home = os.path.expanduser('~')
    filename = '{}/{}'.format(home, file)

    try:
        with open(filename, 'r') as ymlfile:
            return yaml.load(ymlfile)
    except IOError as e:
        raise MissingIMAPCredentailsException('IMAP credentials file: {} is not found'.format(filename))
    except yaml.scanner.ScannerError as e:
        raise InvalidIMAPCredentailsFileException('IMAP credentials file: {} is not valid YAML'.format(filename))


def _create_log_filename(filename):
    if os.name == 'nt':
        logdir = os.path.expandvars(r'${SystemDrive}\cfn\log')
        if not os.path.exists(logdir):
            os.makedirs(logdir)
        return logdir + os.path.sep + filename

    return '/tmp/%s' % filename


def add_stream_log_handler(logger):
    """
    Add StreamHandler to the logger

    :param logger: existing logger
    :return: None
    """
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(logging.Formatter(
            '%(asctime)s %(name)-12s %(levelname)-8s %(message)s'))
    logger.addHandler(handler)


def add_file_log_handler(logger, name):
    """
    Add RotatingFileHandler to the logger

    :param logger: existing logger
    :param name: filename
    :return: None
    """
    handler = RotatingFileHandler(
        backupCount=3, maxBytes=1000000, filename=_create_log_filename('{}.log'.format(name))
    )
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(logging.Formatter(
            '%(asctime)s %(name)-12s %(levelname)-8s %(message)s'))
    logger.addHandler(handler)


def configure_logger(name):
    logger = logging.getLogger(name)
    logger.addHandler(logging.NullHandler())
    logger.setLevel(logging.DEBUG)

    return logger


class ACManagerException(Exception):
    """Basic ACManager exception"""


class MissingIMAPCredentailsException(ACManagerException):
    """Missing IMAP credentails file"""


class InvalidIMAPCredentailsFileException(ACManagerException):
    """IMAP credentails file is not valid YAML"""


class MissingCertificateArgException(ACManagerException):
    """Missing required CLI argument exception"""


class IMAPCredentialFileMissingPropertyException(ACManagerException):
    """IMAP credential file is missing required property"""


class InvalidCertificateJsonFileException(ACManagerException):
    """Missing required CLI argument exception"""


class SMTPConnectionFailedException(ACManagerException):
    """Raised when failed to establish connection with SMTP server"""


class FailedToFetchEmailException(SMTPConnectionFailedException):
    """Raised when failed to fetch email"""


class NoEmailsFoundException(ACManagerException):
    """Raised when emails are not found"""


class EmailBodyUnknownContentType(ACManagerException):
    """Raised when certificate body is not in html format"""


class EmailBodyConfirmLinkIsMissingException(ACManagerException):
    """Raised when certificate email is missing confirmation url"""


class ConfirmPageIsMissingFormException(ACManagerException):
    """Raised when confirm page is missing the actual form"""

UserHeaders = {
    'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36'
}



import os
import io
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

with io.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='acmagent',
    version='1.0.1',
    description='ACM agent - automates ACM certificates',
    long_description=long_description,
    url='https://github.com/b-b3rn4rd/acmagent',
    author='Bernard Baltrusaitis',
    test_suite="tests",
    tests_require=['mock'],
    author_email='bernard@runawaylover.info',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Environment :: Console'
    ],
    keywords='acm aws ssl certificates',
    packages=find_packages(exclude=['docs', 'tests']),
    install_requires=[
        'boto3>=1.4.4',
        'botocore>=1.5.34,<2.0.0',
        'beautifulsoup4>=4.5.3',
        'PyYAML>=3.12',
        'requests>=2.13.0',
    ],
    package_data={
        'acmagent': ['*.json']
    },
    entry_points={
        'console_scripts': [
            'acmagent = acmagent.cli:main'
        ]
    }
)



import imaplib
import email
from bs4 import BeautifulSoup
import requests
import logging
import acmagent
import json

logger = logging.getLogger('acmagent')


class ConfirmCertificate(object):
    """
    Certificate confirmation class, tried to confirm certificate with given id by connecting to IMAP server
    """
    APPROVAL_URL_ID = 'approval_url'
    APPROVAL_FORM_URL = 'https://certificates.amazon.com/approvals'
    EMAIL_FOLDER = 'Inbox'

    def __enter__(self):
        return self

    def __init__(self, imap_credentials):
        try:
            self._server = imap_credentials['server']
            self._username = imap_credentials['username']
            self._password = imap_credentials['password']
        except TypeError as e:
            logger.exception('IMAP credentials file is not well formatted')
            raise acmagent.InvaliIMAPCredentailsFileException('IMAP credentials file is empty or not well formatted')
        except KeyError as e:
            logger.exception('IMAP credentials missing property')
            raise acmagent.IMAPCredentialFileMissingPropertyException(
                'Missing IMAP property "{}", check the credentials file'.format(e.args[0]))

        self._connect_to_imap()

    def _connect_to_imap(self):
        try:
            logger.info('Establishing connection with {} server'.format(self._server))
            self._mail = imaplib.IMAP4_SSL(self._server)
            self._mail.login(self._username, self._password)
        except Exception as e:
            logger.exception('Failed establish IMAP connection: {}'.format(e))
            raise acmagent.SMTPConnectionFailedException('Can\'t login to the "{}" server'.format(self._server))

    @staticmethod
    def _search_query(certificate_id):
        search_query = (
            'UNSEEN',
            'FROM "Amazon Certificates"',
            'BODY "Certificate identifier: {}"'.format(certificate_id)
        )

        return "({})".format(" ".join(search_query))

    def _call_confirm_url(self, url):
        logger.info('Sending GET: {}'.format(url))
        response = requests.get(url, headers=acmagent.UserHeaders)

        try:
            confirm_body = BeautifulSoup(response.content, "html.parser")
            confirm_form = confirm_body.body.find('form').find_all('input')
            payload = {input.get('name'): input.get('value') for input in confirm_form}
            logger.debug('Found confirmation form: {}'.format(json.dumps(payload)))
            return self._call_confirm_form(payload)
        except AttributeError as e:
            logger.exception('Failed to extract confirmation form')
            raise acmagent.ConfirmPageIsMissingFormException('The certificate has been confirmed or the confirmation link: "{}" has expired'.format(url))

    def _call_confirm_form(self, payload):
        logger.info('Sending POST: {} PAYLOAD: {}'.format(ConfirmCertificate.APPROVAL_FORM_URL, json.dumps(payload)))
        response = requests.post(ConfirmCertificate.APPROVAL_FORM_URL, headers=acmagent.UserHeaders, data=payload)

        if not response.ok:
            logger.exception('Failed to submit confirmation form')
            raise acmagent.ACManagerException('An unknown error has occurred while requesting url:"{}"'.format(ConfirmCertificate.APPROVAL_FORM_URL))

        logger.info('Success! The certificate has been confirmed')
        return True

    def _fetch_message(self, message_id):
        type, response = self._mail.fetch(message_id,'(RFC822)')
        email_message = email.message_from_string(response[0][1])
        logger.debug('Opening email: {}'.format(email_message['subject']))
        self._mail.store(message_id, '+FLAGS', '\\Seen')
        logger.debug('Marking email: {} as read'.format(email_message['subject']))
        email_body = ''
        if email_message.is_multipart():
            for multipart in email_message.get_payload():
                if 'text/html' == multipart.get_content_type():
                    email_body = multipart.get_payload(decode=True)
                    break
        if not email_body:
            logger.exception('Email is missing HTML')
            raise acmagent.EmailBodyUnknownContentType('Email "{}" is not in the text/html Content-Type'.format(message_id))

        try:
            logger.debug('Reading email: {} HTML body'.format(email_message['subject']))
            html = BeautifulSoup(email_body, "html.parser")
            approval_url = html.body.find('a', attrs={'id': ConfirmCertificate.APPROVAL_URL_ID}).get('href')
            logger.debug('Found confirmation url: {}'.format(approval_url))
            return self._call_confirm_url(approval_url)
        except AttributeError as e:
            logger.exception('Failed to parse email html')
            raise acmagent.EmailBodyConfirmLinkIsMissingException('Url with "id={}" is not found in the email'.format(ConfirmCertificate.APPROVAL_URL_ID))

    def confirm_certificate(self, certificate_id):
        try:
            self._mail.select(ConfirmCertificate.EMAIL_FOLDER)
            imap_search = ConfirmCertificate._search_query(certificate_id)
            logger.debug('Scan {} folder with {} condition'.format(ConfirmCertificate.EMAIL_FOLDER, imap_search))
            success, messages = self._mail.search(None, imap_search)
            if success == 'OK':
                message_ids = [message_id for message_id in messages[0].split(' ') if message_id]

                if not message_ids:
                    logger.info('Have not found email for requested certificate')
                    raise acmagent.NoEmailsFoundException('Failed to find email for certificate {} in {} folder'.format(
                        certificate_id, ConfirmCertificate.EMAIL_FOLDER))

                logger.debug('Found {} email(s)'.format(len(message_ids)))
                for message_id in message_ids:
                    success = self._fetch_message(message_id)
                    if success:
                        return True
            else:
                logger.exception('Unknown error')
                raise acmagent.ACManagerException('An unknown error has occurred while reading emails, state={}'.format(success))
        except imaplib.IMAP4.error as e:
            if str(e).startswith('FETCH'):
                raise acmagent.FailedToFetchEmailException('Failed to fetch emails')
            elif str(e).startswith('command EXAMINE illegal'):
                raise acmagent.SMTPConnectionFailedException('Can\'t establish connection with "{}" server'.format(self._server))

    def __exit__(self, exc_type, exc_val, exc_tb):
        logger.info('Closing connection with {} server'.format(self._server))
        self._mail.close()
